@@ -1,14 +1,15 @@
 import os
 from collections import OrderedDict
 
-from pycsp3.classes.auxiliary.conditions import Condition
+from pycsp3.classes.auxiliary.conditions import Condition, ConditionInterval, ConditionSet
 from pycsp3.classes.auxiliary.ptypes import TypeVar, TypeCtr, TypeCtrArg, TypeXML, TypeConditionOperator, TypeRank
 from pycsp3.classes.auxiliary.values import IntegerEntity
 from pycsp3.classes.entities import EVarArray, ECtr, TypeNode, Node
 from pycsp3.classes.main.domains import Domain
 from pycsp3.classes.main.variables import Variable, VariableInteger
 from pycsp3.tools.compactor import compact
-from pycsp3.tools.utilities import is_1d_list, matrix_to_string, transitions_to_string, integers_to_string, table_to_string, flatten, is_matrix, error
+from pycsp3.tools.utilities import ANY, is_1d_list, matrix_to_string, transitions_to_string, integers_to_string, table_to_string, flatten, is_matrix, error, \
+    to_ordinary_table
 
 
 class Diffs:
@@ -151,42 +152,106 @@ class ConstraintIntension(Constraint):
 
 class ConstraintExtension(Constraint):
     cache = dict()
+    cache_smart = dict()
 
     @staticmethod
-    def caching(table):
+    def convert_smart_to_ordinary(scope, table):
+        for i, t in enumerate(table):
+            tpl = None
+            for j, v in enumerate(t):
+                if isinstance(v, int) or v == ANY:
+                    if tpl:
+                        tpl.append(v)
+                else:
+                    if isinstance(v, range):
+                        if not tpl:
+                            tpl = list(t[:j])
+                        tpl.append(ConditionInterval(TypeConditionOperator.IN, v.start, v.stop - 1))
+                    elif isinstance(v, (tuple, list, set, frozenset)):
+                        assert all(isinstance(w, int) for w in v)
+                        if not tpl:
+                            tpl = list(t[:j])
+                        tpl.append(ConditionSet(TypeConditionOperator.IN, set(v)))
+                    else:
+                        assert isinstance(v, Condition), "An element of this table should not be of type: " + str(type(v))
+                        if tpl:
+                            tpl.append(v)
+            if tpl:
+                table[i] = tuple(tpl)
+
+        return sorted(list(to_ordinary_table(table, [x.dom for x in scope], keep_any=True)))
+
+    @staticmethod
+    def remove_redundant_tuples(table):
+        tbl = None
+        for i in range(len(table) - 1):
+            if table[i] != table[i + 1]:
+                if tbl:
+                    tbl.append(table[i])
+            else:
+                if not tbl:
+                    tbl = table[:i]
+        if tbl:
+            tbl.append(table[-1])
+            table = tbl
+        return table
+
+    def process_table(self, scope, table):
         if len(table) == 0:
             return None
-        arity = 1 if is_1d_list(table, (int, str)) else len(table[0])
-        h = hash(tuple(table))
-        if h not in ConstraintExtension.cache:
-            if arity > 1:
-                table.sort()
 
-                tbl = None  # we remove redundant tuples, if present
-                for i in range(len(table) - 1):
-                    if table[i] != table[i + 1]:
-                        if tbl:
-                            tbl.append(table[i])
-                    else:
-                        if not tbl:
-                            tbl = table[:i]
-                if tbl:
-                    tbl.append(table[-1])
-                    table = tbl
+        # we compute the hash code of the table
+        try:
+            h = hash(tuple(table))
+        except TypeError:
+            for i, t in enumerate(table):
+                if any(isinstance(v, set) for v in t):
+                    table[i] = tuple(tuple(v) if isinstance(v, set) else v for v in t)
+            h = hash(tuple(table))
 
-                ConstraintExtension.cache[h] = table_to_string(table, parallel=os.name != 'nt')
-            elif isinstance(table[0], int):
-                ConstraintExtension.cache[h] = integers_to_string(table)
+        if len(scope) == 1:  # if arity 1
+            if h not in ConstraintExtension.cache:
+                ConstraintExtension.cache[h] = integers_to_string(table) if isinstance(table[0], int) else " ".join(v for v in sorted(table))
+            return ConstraintExtension.cache[h]
+
+        if h in ConstraintExtension.cache_smart:
+            smart = ConstraintExtension.cache_smart[h]
+        else:
+            smart = any(not (isinstance(v, (int, str)) or v == ANY) for t in table for v in t)  # A parallelisation attempt showed no gain.
+            ConstraintExtension.cache_smart[h] = smart
+        if not smart:
+            if not self.restrictTablesWrtDomains:
+                # we can use caching here
+                if h not in ConstraintExtension.cache:
+                    table.sort()
+                    table = ConstraintExtension.remove_redundant_tuples(table)
+                    ConstraintExtension.cache[h] = table_to_string(table, parallel=os.name != 'nt')
+                return ConstraintExtension.cache[h]
             else:
-                ConstraintExtension.cache[h] = " ".join(v for v in sorted(table))
-        return ConstraintExtension.cache[h]
+                # if ever we would want to use caching, we should include domains when computing hashes
+                table.sort()
+                return table_to_string(table, restricting_domains=[x.dom for x in scope], parallel=os.name != 'nt')
+        else:  # it is smart
+            if self.keepsmartconditions:  # currently, no restriction of tables (wrt domains) in that case
+                self.attributes.append((TypeXML.TYPE, "smart"))
+                if h not in ConstraintExtension.cache:
+                    table = ConstraintExtension.remove_redundant_tuples(table)
+                    ConstraintExtension.cache[h] = table_to_string(table, parallel=os.name != 'nt')
+                return ConstraintExtension.cache[h]
+            else:
+                # if ever we would want to use caching, we should include domains when computing hashes
+                table = ConstraintExtension.convert_smart_to_ordinary(scope, table)
+                table = ConstraintExtension.remove_redundant_tuples(table)
+                return table_to_string(table, restricting_domains=[x.dom for x in scope] if self.restrictTablesWrtDomains else None, parallel=os.name != 'nt')
 
-    def __init__(self, scope, table, positive=True):
+    def __init__(self, scope, table, positive=True, keepsmartconditions=False, restrictTablesWrtDomains=False):
         super().__init__(TypeCtr.EXTENSION)
+        self.keepsmartconditions = keepsmartconditions
+        self.restrictTablesWrtDomains = restrictTablesWrtDomains
         assert is_1d_list(scope, Variable)
         assert len(table) == 0 or (len(scope) == 1 and (is_1d_list(table, int) or is_1d_list(table, str))) or (len(scope) > 1 and len(scope) == len(table[0]))
         self.arg(TypeCtrArg.LIST, scope, content_ordered=True)
-        self.arg(TypeCtrArg.SUPPORTS if positive else TypeCtrArg.CONFLICTS, ConstraintExtension.caching(table), content_compressible=False)
+        self.arg(TypeCtrArg.SUPPORTS if positive else TypeCtrArg.CONFLICTS, self.process_table(scope, table), content_compressible=False)
 
     def close_to(self, other):
         if not self.similar_structure(other):
@@ -510,10 +575,11 @@ class PartialConstraint:  # constraint whose condition has not been given such a
             return other
         if isinstance(other, Node):
             return auxiliary().replace_node(other)
-        assert isinstance(other, (ScalarProduct, PartialConstraint))
+        assert isinstance(other, (ScalarProduct, PartialConstraint)), "Wrong type for " + str(other)
         if isinstance(self.constraint, ConstraintSum) and (isinstance(other, ScalarProduct) or isinstance(other.constraint, ConstraintSum)):
             return other
-        assert isinstance(self.constraint, ConstraintWithCondition) and isinstance(other.constraint, ConstraintWithCondition)
+        assert isinstance(self.constraint, ConstraintWithCondition) and isinstance(other.constraint, ConstraintWithCondition), "Wrong type for " + str(
+            other.constraint) + " or/and " + str(self.constraint)
         return auxiliary().replace_partial_constraint(other)
 
     def _simplify_operation(self, other):
