@@ -1,11 +1,12 @@
 import inspect
+import math
 import types
 from collections import namedtuple
 from itertools import combinations, product, permutations
 
-from pycsp3.classes.auxiliary.conditions import Condition, ConditionValue, ne
-from pycsp3.classes.auxiliary.structures import Automaton, MDD
+from pycsp3.classes.auxiliary.conditions import Condition, ConditionInterval, ConditionSet, lt, le, ge, gt, ne, complement
 from pycsp3.classes.auxiliary.ptypes import TypeOrderedOperator, TypeConditionOperator, TypeVar, TypeCtr, TypeCtrArg, TypeRank
+from pycsp3.classes.auxiliary.structures import Automaton, MDD
 from pycsp3.classes.entities import (
     EVar, EVarArray, ECtr, ECtrs, EToGather, EToSatisfy, EBlock, ESlide, EIfThenElse, EObjective, EAnnotation, AnnEntities, TypeNode, Node)
 from pycsp3.classes.main.annotations import (
@@ -19,13 +20,13 @@ from pycsp3.classes.main.constraints import (
 from pycsp3.classes.main.domains import Domain
 from pycsp3.classes.main.objectives import ObjectiveExpression, ObjectivePartial
 from pycsp3.classes.main.variables import Variable, VariableInteger, VariableSymbolic, NotVariable, NegVariable
+from pycsp3.compiler import default_data
 from pycsp3.dashboard import options
 from pycsp3.tools import curser
-from pycsp3.tools.curser import OpOverrider, ListInt, ListVar, columns
+from pycsp3.tools.curser import OpOverrider, ListInt, ListVar
 from pycsp3.tools.inspector import checkType, extract_declaration_for, comment_and_tags_of, comments_and_tags_of_parameters_of
-from pycsp3.tools.utilities import ANY, flatten, is_containing, is_1d_list, is_1d_tuple, is_matrix, is_square_matrix, transpose, alphabet_positions, all_primes, \
-    integer_scaling
-from pycsp3.compiler import default_data
+from pycsp3.tools.utilities import ANY, flatten, is_1d_list, is_1d_tuple, is_matrix, is_square_matrix, transpose, alphabet_positions, all_primes, \
+    integer_scaling, to_ordinary_table
 
 ''' Global Variables '''
 
@@ -64,7 +65,9 @@ def subvariant(name=None):
 
 
 def Var(term=None, *others, dom=None):
-    assert (term is None) != (dom is None)
+    if term is None and dom is None:
+        dom = Domain(math.inf)
+    assert not (term and dom)
     if term is not None:
         dom = flatten(term, others)
     if not isinstance(dom, Domain):
@@ -91,6 +94,8 @@ def VarArray(*, size, dom, comment=None):
     name = extract_declaration_for("VarArray")
     if comment is None and not isinstance(name, list):
         comment, tags = comment_and_tags_of(function_name="VarArray")
+    else:
+        tags = []
 
     assert isinstance(comment, (str, type(None))), "A comment must be a string (or None). Usually, they are given on plain lines preceding the declaration"
     assert not isinstance(dom, type(lambda: 0)) or len(size) == len(inspect.signature(dom).parameters), \
@@ -113,6 +118,7 @@ def VarArray(*, size, dom, comment=None):
 
 def _bool_interpretation_for_in(left_operand, right_operand, bool_value):
     assert type(bool_value) is bool
+
     if isinstance(left_operand, Variable) and isinstance(right_operand, (tuple, list, set, frozenset, range)) and len(right_operand) == 0:
         return None
     if isinstance(left_operand, (Variable, int, str)) and isinstance(right_operand, (set, frozenset, range)):
@@ -127,6 +133,7 @@ def _bool_interpretation_for_in(left_operand, right_operand, bool_value):
     elif isinstance(left_operand, int) and (is_1d_list(right_operand, Variable) or is_1d_tuple(right_operand, Variable)):
         ctr = Count(right_operand, value=left_operand, condition=(TypeConditionOperator.GE, 1))  # atLeast1 TODO to be replaced by a member/element constraint ?
     else:  #  It is a table constraint
+
         if not hasattr(left_operand, '__iter__'):
             left_operand = [left_operand]
         if not bool_value and len(right_operand) == 0:
@@ -210,6 +217,7 @@ def satisfy(*args):
     no_parameter_satisfy = 0
     nb_parameter_satisfy = len(args)
     comments1, comments2, tags1, tags2 = comments_and_tags_of_parameters_of(function_name="satisfy", args=args)
+
     t = []
     for i, arg in enumerate(args):
         if isinstance(arg, list) and len(arg) > 0:
@@ -220,7 +228,8 @@ def satisfy(*args):
                     if isinstance(l, list) and len(l) > 0 and isinstance(l[0], tuple):
                         arg[j] = _reorder(l)
         no_parameter_satisfy = i
-        assert isinstance(arg, (ECtr, ESlide, Node, bool, list, tuple, type(None), types.GeneratorType)), "non authorized type " + str(type(arg))
+        assert isinstance(arg, (ECtr, ESlide, Node, bool, list, tuple, type(None), types.GeneratorType)), "non authorized type " + str(arg) + " " + str(
+            type(arg))
         if arg is None:
             continue
         arg = list(arg) if isinstance(arg, types.GeneratorType) else arg
@@ -257,40 +266,21 @@ def satisfy(*args):
 ''' Generic Constraints (intension, extension) '''
 
 
-# def is_smart(table):
-#     for t in table:
-#         for v in t:
-#             if isinstance(v, ConditionValue):
-#                 return True
-#     return False
-
-
 def Extension(*, scope, table, positive=True):
     scope = flatten(scope)
     checkType(scope, [Variable])
     assert isinstance(table, list)
-    if len(scope) > 1 and any(isinstance(v, ConditionValue) for t in table for v in t):  # if smart table
-        table = sorted(list(to_ordinary_table(table, [x.dom for x in scope], keep_any=True)))
-
-    checkType(table, [str, int, float])
+    assert len(table) > 0, "A table must be a non-empty list of tuples or integers (or symbols)"
     checkType(positive, bool)
-    assert isinstance(table, list) and len(table) > 0, "A table must be a non-empty list of tuples or integers (or symbols)"
-    assert isinstance(table[0], (tuple, int, str)), "Elements of tables are tuples or integers (or symbols)"
-    # print(table)
-    assert isinstance(table[0], (int, str)) or len(scope) == len(table[0]), (
-        "The length of each tuple must be the same as the arity." + "Maybe a problem with slicing: you must for example write x[i:i+3,0] instead of x[i:i+3][0]")
-    # TODO: this ckecking don't pass on Waterbucket.py, but the xml file is the same that the java version !
-    # if options.checker:
-    #    if not hasattr(Extension, "checked_tables"):
-    #        Extension.checked_tables = set()
-    #    if id(table) not in checked_tables:
-    #        for t in table:
-    #            for i, v in enumerate(t):
-    #                if v not in variables[i].dom:
-    #                    raise ValueError(
-    #                        "Problem: Constraint extension : a value of table is not represented in a domain of a variable : " + str(domainElement))
-    #        checked_tables.add(id(table))
-    return ECtr(ConstraintExtension(scope, table, positive))
+
+    if len(scope) == 1:
+        assert all(isinstance(v, int) if isinstance(scope[0], VariableInteger) else isinstance(v, str) for v in table)
+    else:  # if all(isinstance(x, VariableInteger) for x in scope):
+        for i, t in enumerate(table):
+            assert isinstance(t, tuple)
+            assert len(t) == len(scope), ("The length of each tuple must be the same as the arity."
+                                          + "Maybe a problem with slicing: you must for example write x#[i:i+3,0] instead of x[i:i+3][0]")
+    return ECtr(ConstraintExtension(scope, table, positive, options.keepsmartconditions, options.restricttableswrtdomains))
 
 
 def Intension(node):
@@ -308,10 +298,6 @@ def abs(*args):
     return Node.build(TypeNode.ABS, *args) if len(args) == 1 and isinstance(args[0], (Node, Variable)) else absPython(*args)
 
 
-def expr(operator, *args):
-    return Node.build(operator, *args)
-
-
 def min(*args):
     return Node.build(TypeNode.MIN, *args) if len(args) > 0 and any(isinstance(a, (Node, Variable)) for a in args) else minPython(*args)
 
@@ -340,6 +326,10 @@ def ift(*args):
     return Node.build(TypeNode.IF, *args)
 
 
+def expr(operator, *args):
+    return Node.build(operator, *args)
+
+
 def conjunction(*args):
     return Node.conjunction(*args)
 
@@ -505,12 +495,16 @@ def Sum(term, *others, condition=None):
         return terms, coeffs
 
     terms = list(term) if isinstance(term, types.GeneratorType) else flatten(term, others)
+
     checkType(terms, ([Variable], [Node], [PartialConstraint], [ScalarProduct]))
     auxiliary().replace_partial_constraints(terms)
 
     terms, coeffs = _get_terms_coeffs(terms)
     terms, coeffs = _manage_coeffs(terms, coeffs)
-    # TODO control here some assumptions
+    # if len(terms) == 1 and coeffs is None: return terms
+
+    # TODO control here some assumptions (empty list seems to be possible. See RLFAP)
+
     return _wrapping_by_complete_or_partial_constraint(ConstraintSum(terms, coeffs, Condition.build_condition(condition)))
 
 
@@ -665,18 +659,22 @@ def Clause(term, *others, phases=None):
 
 
 def _optimize(term, minimization):
+    if isinstance(term, PartialConstraint) and isinstance(term.constraint, ConstraintSum) and TypeCtrArg.COEFFS not in term.constraint.arguments:
+        l = term.constraint.arguments[TypeCtrArg.LIST]
+        if len(l.content) == 1:
+            term = l.content[0]  # this was a sum with only one term, so we just consider this term as an expression to be optimized
     if isinstance(term, ScalarProduct):
         term = Sum(term)  # to have a PartialConstraint
     checkType(term, (Variable, Node, PartialConstraint)), "Did you forget to use Sum, e.g., as in Sum(x[i]*3 for i in range(10))"
     satisfy(pc == var for (pc, var) in auxiliary().collected())
     comment, _, tag, _ = comments_and_tags_of_parameters_of(function_name="minimize" if minimization else "maximize", args=[term])
-    type = TypeCtr.MINIMIZE if minimization else TypeCtr.MAXIMIZE
+    otype = TypeCtr.MINIMIZE if minimization else TypeCtr.MAXIMIZE
     if isinstance(term, (Variable, Node)):
         if isinstance(term, Node):
             term.mark_as_used()
-        return EObjective(ObjectiveExpression(type, term)).note(comment[0]).tag(tag[0])  # TODO why only one tag?
+        return EObjective(ObjectiveExpression(otype, term)).note(comment[0]).tag(tag[0])  # TODO why only one tag?
     else:
-        return EObjective(ObjectivePartial(type, term)).note(comment[0]).tag(tag[0])
+        return EObjective(ObjectivePartial(otype, term)).note(comment[0]).tag(tag[0])
 
 
 def minimize(term):
@@ -712,15 +710,8 @@ def annotate(*, decision=None, output=None, varHeuristic=None, valHeuristic=None
 ''' Helpers '''
 
 
-# def column(m, j):
-#     assert is_2d_list(m), "column() can only be called on 2-dimensional lists"
-#     assert all(len(row) > j for row in m), "one row has not at least j+1 elements"
-#     return ListVar(row[j] for row in m)
-#
-#
-# def columns(m):
-#     assert is_matrix(m), "columns() can only be called on matrices"
-#     return ListVar(column(m, j) for j in range(len(m[0])))
+def columns(m):
+    return curser.columns(m)
 
 
 def diagonal_down(m, i=-1, j=-1, check=True):
@@ -762,24 +753,6 @@ def different_values(*args):
     return all(a != b for (a, b) in combinations(args, 2))
 
 
-def to_ordinary_table(table, domains, *, keep_any=False):
-    doms = [range(d) if isinstance(d, int) else d.all_values() if isinstance(d, Domain) else d for d in domains]
-    tbl = set()
-    if keep_any:
-        for t in table:
-            if any(isinstance(v, ConditionValue) for v in t):  # v may be a ConditionValue (with method 'filtering')
-                tbl.update(product(*({v} if isinstance(v, int) or v == ANY else v.filtering(doms[i]) for i, v in enumerate(t))))
-            else:
-                tbl.add(t)
-    else:
-        for t in table:
-            if any(v == ANY or isinstance(v, ConditionValue) for v in t):  # v may be a ConditionValue (with method 'filtering')
-                tbl.update(product(*({v} if isinstance(v, int) else doms[i] if v == ANY else v.filtering(doms[i]) for i, v in enumerate(t))))
-            else:
-                tbl.add(t)
-    return tbl
-
-
 def cp_array(*l):
     if len(l) == 1:
         l = l[0]
@@ -799,4 +772,4 @@ def cp_array(*l):
 
 
 def _pycharm_security():  # for avoiding that imports are removed when reformatting code
-    _ = (permutations, transpose, alphabet_positions, all_primes, integer_scaling, columns, namedtuple, default_data, ne)
+    _ = (permutations, transpose, alphabet_positions, all_primes, integer_scaling, namedtuple, default_data, lt, le, ge, gt, ne, complement)
