@@ -32,8 +32,14 @@ def cursing():
     def _list_mul(self, other):  # for being able to use scalar products
         if is_containing(self, (Variable, Node), check_first_only=True):
             return ScalarProduct(self, other)
+        # if is_containing(self, int) and is_containing(other, (Variable, Node)):
+        #     return ScalarProduct(self, other)
         return list.__mul__(self, other)
 
+    # def _list_rmul(self, other):
+    #     return _list_mul(other, self)
+
+
     def _tuple_contains(self, other):
         if not OpOverrider.activated:
             return self.__contains__(other)
@@ -120,6 +126,7 @@ def cursing():
     curse(dict, "__add__", _dict_add)
     curse(tuple, "__mul__", _tuple_mul)
     curse(list, "__mul__", _list_mul)
+    # curse(list, "__rmul__", _list_rmul)
     curse(tuple, "__contains__", _tuple_contains)
     curse(list, "__contains__", _list_contains)
     curse(set, "__contains__", _set_contains)
@@ -263,40 +270,40 @@ class OpOverrider:
     def __lt__(self, other):
         if self is None or other is None:
             return object.__lt__(self, other)
-        return PartialConstraint.__gt__(other, self) if isinstance(other, PartialConstraint) else Node.build(TypeNode.LT, self, other)
+        return other.__gt__(self) if isinstance(other, (PartialConstraint, ScalarProduct)) else Node.build(TypeNode.LT, self, other)
 
     def __le__(self, other):
         if self is None or other is None:
             return object.__le__(self, other)
-        return PartialConstraint.__ge__(other, self) if isinstance(other, PartialConstraint) else Node.build(TypeNode.LE, self, other)
+        return other.__ge__(self) if isinstance(other, (PartialConstraint, ScalarProduct)) else Node.build(TypeNode.LE, self, other)
 
     def __ge__(self, other):
         if self is None or other is None:
             return object.__ge__(self, other)
         if isinstance(other, int) and other == 1 and isinstance(self, Node) and self.type == TypeNode.DIST:  # we simplify the expression
             return Node.build(TypeNode.NE, self.sons[0], self.sons[1])
-        return PartialConstraint.__le__(other, self) if isinstance(other, PartialConstraint) else Node.build(TypeNode.GE, self, other)
+        return other.__le__(self) if isinstance(other, (PartialConstraint, ScalarProduct)) else Node.build(TypeNode.GE, self, other)
 
     def __gt__(self, other):
         if self is None or other is None:
             return object.__gt__(self, other)
         if isinstance(other, int) and other == 0 and isinstance(self, Node) and self.type == TypeNode.DIST:  # we simplify the expression
             return Node.build(TypeNode.NE, self.sons[0], self.sons[1])
-        return PartialConstraint.__lt__(other, self) if isinstance(other, PartialConstraint) else Node.build(TypeNode.GT, self, other)
+        return other.__lt__(self) if isinstance(other, (PartialConstraint, ScalarProduct)) else Node.build(TypeNode.GT, self, other)
 
     def __eq__(self, other):
         if self is None or other is None:
             return object.__eq__(self, other)
         if isinstance(other, int) and other == 0 and isinstance(self, Node) and self.type == TypeNode.DIST:  # we simplify the expression
             return Node.build(TypeNode.EQ, self.sons[0], self.sons[1])
-        return PartialConstraint.__eq__(other, self) if isinstance(other, PartialConstraint) else Node.build(TypeNode.EQ, self, other)
+        return other.__eq__(self) if isinstance(other, (PartialConstraint, ScalarProduct)) else Node.build(TypeNode.EQ, self, other)
 
     def __ne__(self, other):
         if self is None or other is None:
             return object.__ne__(self, other)
         if isinstance(other, int) and other == 0 and isinstance(self, Node) and self.type == TypeNode.DIST:  # we simplify the expression
             return Node.build(TypeNode.NE, self.sons[0], self.sons[1])
-        return PartialConstraint.__ne__(other, self) if isinstance(other, PartialConstraint) else Node.build(TypeNode.NE, self, other)
+        return other.__ne__(self) if isinstance(other, (PartialConstraint, ScalarProduct)) else Node.build(TypeNode.NE, self, other)
 
     def __or__(self, other):
         return object.__or__(self, other) if None in {self, other} else Node.disjunction(self, other)
@@ -407,7 +414,13 @@ class ListInt(list):
         return ListInt(super().__add__(other))
 
     def __mul__(self, other):
-        if is_containing(other, (Variable, Node)):
+        if is_matrix(self):
+            assert is_matrix(other) and len(self) == len(other) and len(self[0]) == len(other[0])
+            t1, t2 = zip(
+                *[(self[i][j], other[i][j]) for i in range(len(self)) for j in range(len(self[0])) if self[i][j] is not None and other[i][j] is not None])
+            assert is_containing(t2, (Variable, Node))
+            return ScalarProduct(list(t2), list(t1))
+        if is_containing(flatten(other), (Variable, Node)):
             return ScalarProduct(other, self)
         assert is_containing(self, (Variable, Node))
         return ScalarProduct(self, other)
@@ -435,7 +448,13 @@ class ListVar(list):
         return ListVar(super().__add__(other))
 
     def __mul__(self, other):
-        assert is_containing(self, (Variable, Node))
+        if is_matrix(self):
+            assert is_matrix(other) and len(self) == len(other) and len(self[0]) == len(other[0])
+            t1, t2 = zip(
+                *[(self[i][j], other[i][j]) for i in range(len(self)) for j in range(len(self[0])) if self[i][j] is not None and other[i][j] is not None])
+            assert is_containing(t1, (Variable, Node))
+            return ScalarProduct(list(t1), list(t2))
+        assert is_containing(self, (Variable, Node))  # Node possible ?
         return ScalarProduct(self, list(other) if isinstance(other, (tuple, range)) else other)
 
     def __contains__(self, other):
@@ -451,26 +470,48 @@ class ListVar(list):
 
 
 def convert_to_namedtuples(obj):
-    if not hasattr(convert_to_namedtuples, "cnt"):
-        convert_to_namedtuples.cnt = 0
-    if isinstance(obj, tuple):
-        obj = list(obj)  # because if data come from a text file (and not from a JSON file), we have different structures, which leads to problems
-    if isinstance(obj, list):
-        if is_1d_list(obj, int):
-            return ListInt(obj)
-        if is_1d_list(obj, Variable):
-            return ListVar(obj)
-        if is_1d_list(obj, dict):
-            nt = namedtuple("nt" + str(convert_to_namedtuples.cnt), obj[0].keys())
-            convert_to_namedtuples.cnt += 1
-            return [nt(*(convert_to_namedtuples(v) for (k, v) in d.items())) for d in obj]
-        t = [convert_to_namedtuples(v) for v in obj]
-        return ListInt(t) if isinstance(t[0], ListInt) else ListVar(t) if isinstance(t[0], ListVar) else t
-    if isinstance(obj, dict):
-        nt = namedtuple("nt" + str(convert_to_namedtuples.cnt), obj.keys())
-        convert_to_namedtuples.cnt += 1
-        return nt(*(convert_to_namedtuples(v) for (k, v) in obj.items()))
-    return obj
+    def with_only_alphanumeric_keys(obj):
+        if isinstance(obj, dict):
+            if any(not k.isalnum() for k in obj.keys()):
+                return False
+            return all(with_only_alphanumeric_keys(v) for v in obj.values())
+        if isinstance(obj,(int,str)):
+            return True
+        try:
+            iter(obj)
+        except TypeError:  # not iterable
+            return True
+        else:  # iterable
+            return all(with_only_alphanumeric_keys(v) for v in obj)
+
+    def recursive_convert_to_namedtuples(obj):
+        if not hasattr(recursive_convert_to_namedtuples, "cnt"):
+            recursive_convert_to_namedtuples.cnt = 0
+        if isinstance(obj, tuple):
+            obj = list(obj)  # because if data come from a text file (and not from a JSON file), we have different structures, which leads to problems
+        if isinstance(obj, list):
+            if len(obj) == 0:
+                return obj
+            if is_1d_list(obj, int):
+                return ListInt(obj)
+            if is_1d_list(obj, Variable):
+                return ListVar(obj)
+            if is_1d_list(obj, dict):
+                nt = namedtuple("nt" + str(recursive_convert_to_namedtuples.cnt), obj[0].keys())
+                recursive_convert_to_namedtuples.cnt += 1
+                return [nt(*(recursive_convert_to_namedtuples(v) for (k, v) in d.items())) for d in obj]
+            t = [recursive_convert_to_namedtuples(v) for v in obj]
+            return ListInt(t) if isinstance(t[0], ListInt) else ListVar(t) if isinstance(t[0], ListVar) else t
+        if isinstance(obj, dict):
+            nt = namedtuple("nt" + str(recursive_convert_to_namedtuples.cnt), obj.keys())
+            recursive_convert_to_namedtuples.cnt += 1
+            return nt(*(recursive_convert_to_namedtuples(v) for (k, v) in obj.items()))
+        return obj
+
+    if not with_only_alphanumeric_keys(obj):
+        print("\nWarning: some key of some dictionary involved in the data is not alphanumeric, so no conversion to named tuples is performed\n")
+        return obj  # not possible to make the conversion in that case
+    return recursive_convert_to_namedtuples(obj)
 
 
 def is_namedtuple(obj):  # imperfect way of checking, but must be enough for our use (when JSON dumping Compilation.data)
@@ -479,3 +520,11 @@ def is_namedtuple(obj):  # imperfect way of checking, but must be enough for our
         return False
     fields = getattr(t, '_fields', None)
     return isinstance(fields, tuple) and all(type(field) == str for field in fields)
+
+# def to_special_list(t):
+#     assert is_1d_list(t)
+#     if is_containing(t, (Variable, type(None))):
+#         return ListVar(t)
+#     if is_containing(t, (int, type(None))):
+#         return ListInt(t)
+#     return t
